import * as Rx from 'rxjs';

export const State = {
  OK: 'OK',
  ERROR: 'ERROR'
};

export class QEntryRun {
  public started: Date;
  public _completed: Date;
  public _error: any;

  constructor() {
    this.started = new Date();
  }

  public error(err: any): void {
    this._error = err;
  }

  public completed(): void {
    this._completed = new Date();
  }
}

export class QEntry<T> {
  public created: Date;
  public runs: QEntryRun[];
  public input: Rx.Observable<T>;
  public output: Rx.Subject<T>;

  constructor(input: Rx.Observable<T>, output: Rx.Subject<T>) {
    this.runs = [];
    this.input = input;
    this.output = output;
    this.created = new Date();
  }

  public isCompleted(): boolean {
    return !!this.runs.find(r => !!r.completed);
  }

  public isWaiting(): boolean {
    // console.log(this.runs);
    return this.runs.length == 0 || !!this.runs.find(qer => !qer.completed);
  }

  public newQEntryRun(): QEntryRun {
    const qer = new QEntryRun();
    this.runs.push(qer);
    return qer;
  }
}

export interface QConfig {
  taskTimer?: number;
  reclaimTimeout?: number;
  retryWaitTime?: number;
  maxExecuteCnt?: number;
}

export class LogMsg {
  public readonly level: string;
  public readonly parts: any[];
  constructor(level: string, parts: any[]) {
    this.level = level;
    this.parts = parts;
  }
}

export interface QWorker<T> {
  (input: Rx.Observable<T>, output: Rx.Subject<T>): void;
}

export class WorkerState<T> {
  private running: QEntryRun;
  private output: Rx.Subject<T>;
  private readonly qWorker: QWorker<T>;
  private readonly q: Queue<T>;

  constructor(q: Queue<T>, qw: QWorker<T>) {
    this.q = q;
    this.qWorker = qw;
    this.running = null;
  }
  public isFree(): boolean {
    return !this.running;
  }

  public run(qe: QEntry<any>): void {
    this.running = qe.newQEntryRun();
    this.output = new Rx.Subject();
    this.output.subscribe(
      (a) => {
        // console.log('WorkerState:next', a);
        qe.output.next(a);
      },
      (err) => {
        // console.log('WorkerState:error', err);
        this.running.error(err);
        this.running = null;
        qe.output.error(err);
      },
      () => {
        // console.log('WorkerState:completed');
        this.running.completed();
        this.running = null;
        qe.output.complete();
        this.q.processMemoryQ();
      }
    );
    this.qWorker(qe.input, this.output);
  }
}

export type QueueSubject<T> = Rx.Subject<Queue<T>|LogMsg>;
export type QueueObserver<T> = Rx.Observer<Queue<T>|LogMsg>;
export type QueueObservable<T> = Rx.Observable<Queue<T>|LogMsg>;

export class Logger<T> {
  private readonly upStream: QueueSubject<T>;
  private readonly base: string[];
  constructor(upStream: QueueSubject<T>, qid: number) {
    this.upStream = upStream;
    this.base = [`Q[${qid}]:`];
  }
  public info(...args: any[]): void {
    this.upStream.next(new LogMsg('info', this.base.concat(args)));
  }
  public error(...args: any[]): void {
    this.upStream.next(new LogMsg('error', this.base.concat(args)));
  }
  public debug(...args: any[]): void {
    this.upStream.next(new LogMsg('debug', this.base.concat(args)));
  }
}

export class Queue<T> {
  // public q: Rx.Subject<QEntry<T>>;
  // public readonly deadLetter: Rx.Subject<QEntry<T>>;
  public qEntries: QEntry<any>[] = [];
  // public inProcess: boolean;
  public qTask: number;
  public reclaimTimeout: number;
  public taskTimer: number;
  public retryWaitTime: number;
  public maxExecuteCnt: number;
  public readonly workers: WorkerState<T>[];
  private readonly lengthSubject: Rx.Subject<number>;
  // private readonly obs: QueueObserver<T>;
  public readonly logger: Logger<T>;

  constructor(obs: QueueSubject<T>, argv: QConfig, id: number) {
    // this.inProcess = false;
    // this.obs = obs;
    this.logger = new Logger(obs, id);
    this.lengthSubject = new Rx.Subject();
    this.taskTimer = argv.taskTimer || 500;
    this.reclaimTimeout = argv.reclaimTimeout || 10000;
    this.retryWaitTime = argv.retryWaitTime || 1000;
    this.maxExecuteCnt = argv.maxExecuteCnt || 10;
    this.workers = [];
    // this.q = new Rx.Subject<QEntry<T>>();
    // this.q.subscribe(this.action.bind(this));
    // this.deadLetter = new Rx.Subject<QEntry<T>>();
    this.qTask = setInterval(this.processMemoryQ.bind(this), this.taskTimer);
  }

  public addWorker(qw: QWorker<T>): Queue<T> {
    this.workers.push(new WorkerState(this, qw));
    return this;
  }

  public length(): Rx.Subject<number> {
    return this.lengthSubject;
  }

  public stop(): Rx.Observable<void> {
    this.logger.info('run stop');
    // console.log('run stop');
    return Rx.Observable.create((observer: Rx.Observer<void>) => {
      this.logger.info('queue stop');
      let forceStopAfter = 3;
      const action = () => {
        if (forceStopAfter <= 0 || this.qEntries.length == 0) {
          clearInterval(this.qTask);
          this.logger.info('Q Task stopped');
          observer.next(null);
          observer.complete();
        } else {
          --forceStopAfter;
          this.logger.info('waiting for Q Task to stop', this.qEntries.length, forceStopAfter);
          setTimeout(action, this.taskTimer);
        }
      };
      action();
    });
  }

  // private action(qe: QEntry<T>): void {
  //   this.logger.debug('action:', qe);
  //   // qe.running = new Date();
  //   // ++qe.executeCnt;
  //   // qe.task.subscribe((_qe: T) => {
  //   //   qe.running = null;
  //   //   qe.completed = new Date();
  //   //   this.logger.debug('action:completed:', qe);
  //   // }, (error: any) => {
  //   //   qe.running = null;
  //   //   qe.retryAt = new Date((new Date()).getTime() + this.retryWaitTime);
  //   //   this.logger.debug('action:error:', qe.created.getTime(),
  //   //                                      qe.retryAt.getTime(),
  //   //                                      qe);
  //   // });
  // }

  public processMemoryQ(): void {
    const freeWorker = this.workers.find(w => w.isFree());
    const qentry = this.qEntries.find(q => q.isWaiting());
    if (freeWorker && qentry) {
      // this.logger.info('processMemoryQ:run');
      freeWorker.run(qentry);
    }
    const prevLen = this.qEntries.length;
    this.qEntries = this.qEntries.filter(a => !a.isCompleted());
    if (prevLen != this.qEntries.length) {
      this.lengthSubject.next(this.qEntries.length);
    }
    // this.qEntries = this.qEntries.filter((qe) => {
    //   if (qe.completed) {
    //     this.logger.debug('completed remove', qe);
    //     return false;
    //   } else if (qe.executeCnt >= this.maxExecuteCnt) {
    //     this.logger.error('drop queue entry:', qe);
    //     this.deadLetter.next(qe);
    //     return false;
    //   } else if (qe.retryAt) {
    //     if (now >= qe.retryAt.getTime()) {
    //       qe.retryAt = null;
    //       this.logger.debug('retry queue entry:', qe);
    //       this.q.next(qe);
    //     }
    //     return true;
    //   } else if (!qe.running) {
    //     this.logger.debug('trigger action:', qe);
    //     this.q.next(qe);
    //   } else if ((qe.running && qe.executeCnt < this.maxExecuteCnt &&
    //     (now - qe.running.getTime()) >= this.reclaimTimeout)) {
    //     qe.running = null;
    //     this.logger.debug('retry execution time entry:', qe);
    //     this.q.next(qe);
    //   }
    //   return true;
    // });
  }

  public push<Y>(action: Rx.Observable<Y>, completed: Rx.Subject<Y>): void {
    this.qEntries.push(new QEntry<Y>(action, completed));
    this.lengthSubject.next(this.qEntries.length);
    this.processMemoryQ();
  }

}

let queueId = ~~(0x1000000 * Math.random());
export function start<T>(argv: any): QueueObservable<T> {
  return Rx.Observable.create((obs: QueueSubject<T>) => {
    const q = new Queue<T>(obs, argv, queueId++);
    q.logger.info('Queue Started');
    obs.next(q);
  });
}

export default Queue;
